const coral = @import("coral");

const ext = @import("./ext.zig");

pub const file = @import("./file.zig");

pub const heap = @import("./heap.zig");

pub const script = @import("./script.zig");

pub const GraphicsLoop = opaque {
	const Internal = struct {
		window: *ext.SDL_Window,
		renderer: *ext.SDL_Renderer,
		loaders: GraphicsLoaderList,
		updaters: GraphicsUpdaterList,
		target_frame_time: f64,

		keys: struct {
			pressed: [ext.SDL_NUM_SCANCODES]bool = [_]bool{false} ** ext.SDL_NUM_SCANCODES,
			released: [ext.SDL_NUM_SCANCODES]bool = [_]bool{false} ** ext.SDL_NUM_SCANCODES,
			held: [ext.SDL_NUM_SCANCODES]bool = [_]bool{false} ** ext.SDL_NUM_SCANCODES,
		},

		const GraphicsLoaderList = coral.list.Stack(GraphicsLoader);

		const GraphicsUpdaterList = coral.list.Stack(GraphicsUpdater);

		fn deinit(self: *Internal) void {
			ext.SDL_DestroyWindow(self.window);
			ext.SDL_DestroyRenderer(self.renderer);
		}

		fn init(allocator: coral.io.Allocator) coral.io.AllocationError!Internal {
			const window = create: {
				const height = 480;
				const width = 640;
				const pos = ext.SDL_WINDOWPOS_CENTERED;
				const flags = ext.SDL_WINDOW_HIDDEN;

				break: create ext.SDL_CreateWindow("", pos, pos, width, height, flags) orelse {
					return error.OutOfMemory;
				};
			};

			errdefer ext.SDL_DestroyWindow(window);

			const renderer = create: {
				const default_driver_index = -1;
				const flags = ext.SDL_RENDERER_ACCELERATED;

				break: create ext.SDL_CreateRenderer(window, default_driver_index, flags) orelse {
					return error.OutOfMemory;
				};
			};

			errdefer ext.SDL_DestroyRenderer(renderer);

			return .{
				.loaders = GraphicsLoaderList.init(allocator),
				.updaters = GraphicsUpdaterList.init(allocator),
				.target_frame_time = 1.0 / 60.0,
				.window = window,
				.renderer = renderer,
				.keys = .{},
			};
		}

		fn run(self: *Internal) void {
			var ticks_previous = ext.SDL_GetTicks64();
			var accumulated_time = @as(f64, 0);

			for (self.loaders.values) |loader| {
				loader.invoke(@ptrCast(self));
			}

			ext.SDL_ShowWindow(self.window);

			running_loop: while (true) {
				const ticks_current = ext.SDL_GetTicks64();
				const milliseconds_per_second = 1000.0;
				const delta_time = @as(f64, @floatFromInt(ticks_current - ticks_previous)) / milliseconds_per_second;

				ticks_previous = ticks_current;
				accumulated_time += delta_time;

				while (accumulated_time >= self.target_frame_time) : (accumulated_time -= self.target_frame_time) {
					var event = @as(ext.SDL_Event, undefined);

					coral.io.zero(@ptrCast(&self.keys.pressed));
					coral.io.zero(@ptrCast(&self.keys.released));

					while (ext.SDL_PollEvent(&event) != 0) {
						switch (event.type) {
							ext.SDL_QUIT => break: running_loop,

							ext.SDL_KEYDOWN => {
								self.keys.pressed[event.key.keysym.scancode] = true;
								self.keys.held[event.key.keysym.scancode] = true;
							},

							ext.SDL_KEYUP => {
								self.keys.released[event.key.keysym.scancode] = true;
								self.keys.held[event.key.keysym.scancode] = false;
							},

							else => {},
						}
					}

					for (self.updaters.values) |updater| {
						updater.invoke(.{
							.loop = @ptrCast(self),
							.delta_time = self.target_frame_time,
						});
					}
				}

				_ = ext.SDL_SetRenderDrawColor(self.renderer, 0, 0, 0, 255);
				_ = ext.SDL_RenderClear(self.renderer);
				_ = ext.SDL_RenderPresent(self.renderer);

				ext.SDL_Delay(1);
			}
		}

		fn use(self: *Internal, middleware: GraphicsMiddleware) coral.io.AllocationError!void {
			try switch (middleware) {
				.loader => |loader| self.loaders.push_one(loader),
				.updater => |updater| self.updaters.push_one(updater),
			};
		}
	};

	fn internal(self: *GraphicsLoop) *Internal {
		return @ptrCast(@alignCast(self));
	}

	pub fn is_key_pressed(self: *GraphicsLoop, key: Key) bool {
		return self.internal().keys.pressed[@intFromEnum(key)];
	}

	pub fn set_resolution(self: *GraphicsLoop, width: u16, height: u16) callconv(.C) void {
		const pos = ext.SDL_WINDOWPOS_CENTERED;
		const internal_state = self.internal();

		ext.SDL_SetWindowSize(internal_state.window, width, height);
		ext.SDL_SetWindowPosition(internal_state.window, pos, pos);
		ext.SDL_ShowWindow(internal_state.window);
	}

	pub fn set_tick_rate(self: *GraphicsLoop, tick_rate: f64) callconv(.C) void {
		self.internal().target_frame_time = 1.0 / tick_rate;
	}

	pub fn set_title(self: *GraphicsLoop, title: coral.utf8.SmallString) callconv(.C) void {
		ext.SDL_SetWindowTitle(self.internal().window, title.to_units());
	}
};

pub const GraphicsUpdate = struct {
	loop: *GraphicsLoop,
	delta_time: f64,
};

pub const GraphicsMiddleware = union (enum) {
	loader: GraphicsLoader,
	updater: GraphicsUpdater,
};

pub const GraphicsLoader = coral.io.Generator(void, *GraphicsLoop);

pub const GraphicsUpdater = coral.io.Generator(void, GraphicsUpdate);

pub const Key = enum (usize) {
	f5 = ext.SDL_SCANCODE_F5,
};

fn last_sdl_error() [:0]const u8 {
	return coral.io.slice_sentineled(@as(u8, 0), @as([*:0]const u8, @ptrCast(ext.SDL_GetError())));
}

pub fn log_info(message: []const coral.io.Byte) void {
	ext.SDL_LogInfo(
		ext.SDL_LOG_CATEGORY_APPLICATION,
		"%.*s",
		coral.math.clamped_cast(@typeInfo(c_int).Int, message.len),
		message.ptr,
	);
}

pub fn log_warn(message: []const coral.io.Byte) void {
	ext.SDL_LogWarn(
		ext.SDL_LOG_CATEGORY_APPLICATION,
		"%.*s",
		coral.math.clamped_cast(@typeInfo(c_int).Int, message.len),
		message.ptr,
	);
}

pub fn log_fail(message: []const coral.io.Byte) void {
	ext.SDL_LogError(
		ext.SDL_LOG_CATEGORY_APPLICATION,
		"%.*s",
		coral.math.clamped_cast(@typeInfo(c_int).Int, message.len),
		message.ptr,
	);
}

pub fn start_graphics(middlewares: []const GraphicsMiddleware) void {
	const init_flags = ext.SDL_INIT_EVERYTHING;

	if (ext.SDL_WasInit(init_flags) == init_flags) {
		return log_fail("a loop is already running");
	}

	if (ext.SDL_Init(init_flags) != 0) {
		return log_fail(last_sdl_error());
	}

	defer ext.SDL_Quit();

	var loop_internal = GraphicsLoop.Internal.init(heap.allocator) catch {
		return log_fail(last_sdl_error());
	};

	defer loop_internal.deinit();

	for (middlewares) |middleware| {
		loop_internal.use(middleware) catch |allocation_error| {
			return log_fail(switch (allocation_error) {
				error.OutOfMemory => "could not use middleware: out of memory",
			});
		};
	}

	loop_internal.run();
}
